
class PathFinder(object):
    def __init__(self, ontology=None):
        """
        Arguments
        ---------

        ontology: Ontology
            An ontology used in pathfinding
        """
        self.ontology = ontology
        
    def find_paths(start=None, end=None):
        
